from __future__ import annotations

import contextlib
from dataclasses import dataclass, field
import hmac
import itertools
import json
import logging
import secrets
import time
from typing import Any, Iterator, Optional, Union
from urllib.parse import quote

from requests import Session

from . import requests as _r
from .errors import AskfmApiError, OperationError, RequestError, SessionError

# === Defaults ===
DEFAULT_HEADERS = {
    "Accept": "application/json; charset=utf-8",
    "Accept-Encoding": "identity",
    "User-Agent": "Dalvik/2.1.0 (Linux; U; Android 6.0.1; GT-N7100 Build/MOB30R)",
    "X-Api-Version": "1.18",
    "X-Client-Type": "android_4.67.1",
}
DEFAULT_HOST = "api.ask.fm:443"
DEFAULT_LIMIT = 50

# === Types ===
ReqParams = dict[str, Any]
StrParams = dict[str, str]
Response = Any
Auth = tuple[str, str]  # login, password


@dataclass
class Request:
    method: str
    path: str
    params: ReqParams = field(default_factory=dict)
    name: Optional[str] = None
    unwrap_key: Optional[str] = None
    paginated: bool = False
    item_id_key: Optional[str] = None


class AskfmApi:
    api_key: bytes
    auto_refresh_session: bool
    device_id: str
    auth: Optional[Auth]
    headers: dict[str, str]

    logged_in: bool = False
    sess: Session
    rt: str = "1"

    _access_token: Optional[str] = None
    _host: str

    @classmethod
    def random_device_id(cls):
        return secrets.token_hex(8)

    def __init__(
        self,
        api_key: Union[str, bytes],
        *,
        auto_refresh_session: bool = True,
        device_id: Optional[str] = None,
        access_token: Optional[str] = None,
        logged_in: bool = False,
        auth: Optional[Auth] = None,
        host: str = DEFAULT_HOST,
        headers: dict[str, str] = DEFAULT_HEADERS,
    ) -> None:
        if isinstance(api_key, str):
            api_key = bytes(api_key, encoding="ascii")
        self.api_key = api_key
        self.auto_refresh_session = auto_refresh_session
        self.device_id = device_id or AskfmApi.random_device_id()
        self.auth = auth

        self.sess = Session()
        self.sess.headers.update(headers)
        self.host = host

        if access_token is not None:
            self.access_token = access_token
            self.logged_in = logged_in
            # In theory, we could detect this automatically.
            # Seems that authenticated tokens start with ".3" and anon tokens with ".5".
        elif logged_in:
            raise TypeError("passed `logged_in` without `access_token`")
        elif auto_refresh_session:
            self.refresh_session()

    @property
    def access_token(self) -> Optional[str]:
        return self._access_token

    @access_token.setter
    def access_token(self, token: Optional[str]) -> None:
        self._access_token = token
        self.sess.headers["X-Access-Token"] = token  # type: ignore

    @property
    def host(self) -> str:
        return self._host

    @host.setter
    def host(self, host: str) -> None:
        self._host = host
        self.sess.headers["Host"] = host

    def log_in(self, username: str, password: str) -> Response:
        return self.refresh_session((username, password))

    def refresh_session(self, auth: Optional[Auth] = None) -> Optional[Response]:
        with self.disable_auto_refresh():  # guard against recursive refreshes
            # get initial anon token
            self.access_token = self.request(_r.get_access_token(self.device_id))
            self.logged_in = False
            if auth := (auth or self.auth):
                # get user token
                res = self.request(_r.log_in(auth[0], auth[1], self.device_id))
                self.access_token = res["accessToken"]
                self.logged_in = True
                return res["user"]
        return None

    def check_session(self) -> bool:
        with self.disable_auto_refresh():
            with contextlib.suppress(RequestError, OperationError):
                self.request(_r.fetch_my_profile())
                self.logged_in = True
                return True
            self.logged_in = False
            self.access_token = None
            return False

    @contextlib.contextmanager
    def disable_auto_refresh(self):
        old_val = self.auto_refresh_session
        self.auto_refresh_session = False
        try:
            yield
        finally:
            self.auto_refresh_session = old_val

    def request(
        self,
        req: Request,
        *,
        unwrap: bool = True,
        offset: int = 0,
        limit: int = DEFAULT_LIMIT,
    ) -> Response:
        if not self.access_token and self.auto_refresh_session:
            self.refresh_session()

        if req.paginated:
            params = {"offset": offset, "limit": limit, **req.params}
        else:
            params = req.params

        for i in itertools.count():
            res = self.request_raw(req.method, req.path, params)
            if "error" not in res:
                break
            error = AskfmApiError.from_response(res)
            if not self.handle_error(error, req, i):
                raise error

        if unwrap and req.unwrap_key:
            res = res[req.unwrap_key]
        return res

    def handle_error(self, error: AskfmApiError, req: Request, attempt_no: int) -> bool:
        """Invalidate/refresh the session and return whether we should retry the request."""
        if isinstance(error, SessionError):
            self.logged_in = False
            self.access_token = None
            if attempt_no == 1 and self.auto_refresh_session:
                self.refresh_session()
                return True
        return False

    def request_iter(
        self,
        req: Request,
        *,
        offset: int = 0,
        page_limit: int = DEFAULT_LIMIT,
    ) -> Iterator[Response]:
        if not req.paginated or not req.unwrap_key:
            raise TypeError("Cannot iterate non-paginated request")

        while True:
            res = self.request(req, offset=offset, limit=page_limit, unwrap=False)
            items = res[req.unwrap_key]
            if not items:  # strangely, API always returns hasMore=True
                break
            yield from items
            if not res["incomplete"] and not res["hasMore"]:
                break
            offset += len(items)

    def request_raw(
        self, method: str, path: str, params: Optional[ReqParams] = None
    ) -> Response:
        params = params or {}
        method = method.upper()
        has_body = method in ["POST", "PUT"]
        url = "https://" + self.host + path

        params = {"rt": self.rt, "ts": int(time.time()), **params}
        params = self.normalize_params(params)
        if has_body:
            params = {"json": json.dumps(params, sort_keys=True, separators=(",", ":"))}
        signature = f"HMAC {self.get_signature(method, path, params)}"

        res = self.sess.request(
            method,
            url,
            data=params if has_body else None,
            params=params if not has_body else None,
            headers={"Authorization": signature},
        )

        if "X-Next-Token" in res.headers:
            self.rt = res.headers["X-Next-Token"]

        return res.json()

    def normalize_params(self, params: ReqParams) -> ReqParams:
        result = {}
        for k, v in params.items():
            if v is None:
                continue
            if isinstance(v, (int, bool)):
                v = str(v).lower()
            result[k] = v
        return result

    def get_signature(self, method: str, path: str, params: StrParams) -> str:
        quoted = [key + "%" + quote(val, safe="!'()~") for (key, val) in params.items()]
        msg = "%".join(sorted(quoted))
        msg = "%".join([method.upper(), self.host, path, msg])

        hmac_ = hmac.new(self.api_key, msg.encode(), "sha1")
        return hmac_.hexdigest()



from __future__ import annotations

from typing import Type

import askfm_api


class AskfmApiError(Exception):
    """Base class for all errors in askfm_api."""

    def __init__(self, response: askfm_api.Response) -> None:
        code = response["error"]
        super().__init__(code)
        self.response = response
        self.code = code

    @staticmethod
    def from_response(response: askfm_api.Response) -> AskfmApiError:
        code = response["error"]
        cls = ERROR_CODE_MAP.get(code, UnknownError)
        return cls(response)


class UnknownError(AskfmApiError):
    """Error for which we don't know it's category."""


class TryAgainError(AskfmApiError):
    """The server asks us to retry."""


class RequestError(AskfmApiError):
    """Incorrect API usage, network problems, banned IP, and so on."""


class OperationError(AskfmApiError):
    """Error performing the requested operation."""


class SessionError(RequestError):
    """Error that should go away after refreshing the session token."""


# I tried my best at guessing the meanings of these codes, but didn't check it.
ERROR_CODE_MAP: dict[str, Type[AskfmApiError]] = {
    "account_banned": OperationError,  # account_banned
    "account_disabled": OperationError,  # account_disabled
    "already_exist": OperationError,  # pinned_answer_already_added
    "already_taken": OperationError,  # already_taken
    "argument_missing": RequestError,  # argument_missing
    "bad_email": OperationError,  # wrong_email_format
    "bad_login": OperationError,  # username_invalid_chars
    "cancelled": AskfmApiError,  # action_cancelled
    "existing_login": OperationError,  # username_taken
    "file_size_error": OperationError,  # file_size_error
    "file_upload_error": RequestError,  # file_upload_error
    "generic_error": AskfmApiError,  # generic_error
    "hashtag_already_exists": OperationError,  # interest_already_exists
    "hashtag_too_long": OperationError,  # interest_too_long
    "illegal_action": RequestError,  # illegal_action
    "incorrect_password": OperationError,  # incorrect_password
    "invalid_access_token": SessionError,  # invalid_access_token
    "invalid_auth": SessionError,  # invalid_auth
    "invalid_data": RequestError,  # invalid_data
    "invalid_email": OperationError,  # wrong_email_format
    "invalid_login": OperationError,  # username_invalid_chars
    "invalid_name": OperationError,  # invalid_name
    "invalid_password": OperationError,  # incorrect_password
    "invalid_request_token": SessionError,  # invalid_request_token
    "invalid_signature": RequestError,  # invalid_signature
    "invalid_user_credentials": OperationError,  # invalid_user_credentials
    "ip_banned": RequestError,  # ip_banned
    "limit_exceeded": OperationError,  # pinned_answers_limit_exceeded
    "network_error": RequestError,  # network_error
    "not_allowed": RequestError,  # not_allowed
    "session_expired": SessionError,  # session_expired
    "session_invalid": SessionError,  # session_invalid
    "social_error": OperationError,  # generic_social
    "something_went_wrong": AskfmApiError,  # something_went_wrong
    "spam_login": OperationError,  # username_forbidden
    "too_many_hashtags": OperationError,  # too_many_interests
    "try_again": TryAgainError,  # try_again
    "unable_to_follow": OperationError,  # unable_to_follow
    "user_not_found": OperationError,  # user_not_found
    "weak_password": OperationError,  # weak_password
    "wrong_format": OperationError,  # wrong_format
    "wrong_hashtag_format": OperationError,  # invalid_interest
}



from __future__ import annotations

import functools
from typing import Optional, Union

import askfm_api

IdType = Union[str, int]

# Note: typechecking this mess seems impossible.
# See https://github.com/python/mypy/issues/3157#issue-221120895, section "Messing with the return type".
def make_req(method, path, **req_kwargs):
    def decorator(func):
        name = func.__name__

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            params = func(*args, **kwargs)
            return askfm_api.Request(method, path, params, name=name, **req_kwargs)

        return wrapper

    return decorator


# === My profile ===


@make_req("GET", "/my/profile", unwrap_key="profile")
def fetch_my_profile():
    return {}


@make_req(
    "GET",
    "/notifications",
    unwrap_key="notifications_ext",
    paginated=True,
    item_id_key="entityId",
)
def fetch_notifs(filter: str = ""):
    return {"type": filter}


@make_req("PUT", "/notifications/mark_read")
def mark_notifs_as_read(filter: str = ""):
    return {"type": filter}


@make_req("POST", "/users/hashtags", unwrap_key="hashtag")
def add_hashtag(hashtag: str):
    return {"hashtag": hashtag}


@make_req("DELETE", "/users/hashtags")
def delete_hashtag(hashtag: str):
    return {"hashtag": hashtag}


# === My questions ===


@make_req(
    "GET", "/my/questions", unwrap_key="questions", paginated=True, item_id_key="qid"
)
def fetch_questions(*, skip_shoutouts: bool = False):
    return {"skip_shoutouts": skip_shoutouts}


@make_req("DELETE", "/my/questions")
def delete_question(question_type: str, question_id: IdType):
    return {
        "qid": question_id,
        "type": question_type,
    }


@make_req("POST", "/my/questions/answer")
def post_answer(question_type: str, question_id: IdType, text: str):
    return {
        "qid": question_id,
        "type": question_type,
        "answer": {
            "type": "text",
            "body": text,
        },
    }


# === Users ===


@make_req(
    "GET",
    "/users/hashtags/search",
    unwrap_key="users",
    paginated=True,
    item_id_key="uid",
)
def search_users_by_hashtag(*hashtags: str):
    return {"hashtags": ",".join(hashtags)}


@make_req("GET", "/users/details", unwrap_key="user")
def fetch_profile(username: str):
    return {"uid": username}


@make_req("POST", "/users/questions")
def send_question(users: list[str], text: str, anon: bool = False):
    return {
        "users": users,
        "question": {
            "type": "anonymous" if anon else "user",
            "body": text,
        },
    }


# === Report/block ===


@make_req("POST", "/report/answer")
def report_answer(question_id: IdType, reason: Optional[str] = None):
    return {"qid": question_id, "reason": reason}


@make_req("POST", "/report/question")
def report_question(
    question_id: IdType, reason: Optional[str] = None, should_block: bool = False
):
    return {"qid": question_id, "reason": reason, "block": should_block}


@make_req("POST", "/report/user")
def report_user(
    username: str, reason: Optional[str] = None, should_block: bool = False
):
    return {"uid": username, "reason": reason, "block": should_block}


# === Util ===


@make_req("GET", "/token", unwrap_key="accessToken")
def get_access_token(device_id: IdType):
    return {"did": device_id}


@make_req("POST", "/authorize")
def log_in(username: str, password: str, device_id: IdType):
    return {
        "uid": username,
        "pass": password,
        "did": device_id,
        "guid": device_id,
    }
